@@ -262,7 +262,7 @@ class CommandContext:
                         os.path.join(self.debian_dir.fs_path, "control"),
                     )
                 with debian_control.open() as fd:
-                    source_package, binary_packages = (
+                    _, source_package, binary_packages = (
                         self.dctrl_parser.parse_source_debian_control(
                             fd,
                         )
