@@ -1,14 +1,15 @@
+import dataclasses
 import re
 import textwrap
 from typing import (
     Union,
     Sequence,
     Tuple,
-    Iterator,
     Optional,
-    Iterable,
     Mapping,
     List,
+    Dict,
+    Any,
 )
 
 from lsprotocol.types import (
@@ -16,14 +17,11 @@ from lsprotocol.types import (
     Range,
     Diagnostic,
     Position,
-    DidOpenTextDocumentParams,
-    DidChangeTextDocumentParams,
     FoldingRange,
     FoldingRangeParams,
     CompletionItem,
     CompletionList,
     CompletionParams,
-    TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL,
     DiagnosticRelatedInformation,
     Location,
     HoverParams,
@@ -31,14 +29,22 @@ from lsprotocol.types import (
     TEXT_DOCUMENT_CODE_ACTION,
     SemanticTokens,
     SemanticTokensParams,
+    WillSaveTextDocumentParams,
+    TextEdit,
+    DocumentFormattingParams,
 )
 
 from debputy.linting.lint_util import LintState
+from debputy.lsp.debputy_ls import DebputyLanguageServer
+from debputy.lsp.diagnostics import DiagnosticData
 from debputy.lsp.lsp_debian_control_reference_data import (
     DctrlKnownField,
     BINARY_FIELDS,
     SOURCE_FIELDS,
     DctrlFileMetadata,
+    package_name_to_section,
+    all_package_relationship_fields,
+    extract_first_value_and_position,
 )
 from debputy.lsp.lsp_features import (
     lint_diagnostics,
@@ -47,17 +53,23 @@ from debputy.lsp.lsp_features import (
     lsp_standard_handler,
     lsp_folding_ranges,
     lsp_semantic_tokens_full,
+    lsp_will_save_wait_until,
+    lsp_format_document,
 )
 from debputy.lsp.lsp_generic_deb822 import (
     deb822_completer,
     deb822_hover,
     deb822_folding_ranges,
     deb822_semantic_tokens_full,
+    deb822_token_iter,
+    deb822_format_file,
 )
 from debputy.lsp.quickfixes import (
     propose_remove_line_quick_fix,
     range_compatible_with_remove_line_fix,
     propose_correct_text_quick_fix,
+    propose_insert_text_on_line_after_diagnostic_quick_fix,
+    propose_remove_range_quick_fix,
 )
 from debputy.lsp.spellchecking import default_spellchecker
 from debputy.lsp.text_util import (
@@ -67,18 +79,15 @@ from debputy.lsp.text_util import (
     te_range_to_lsp,
 )
 from debputy.lsp.vendoring._deb822_repro import (
-    parse_deb822_file,
     Deb822FileElement,
     Deb822ParagraphElement,
 )
 from debputy.lsp.vendoring._deb822_repro.parsing import (
     Deb822KeyValuePairElement,
     LIST_SPACE_SEPARATED_INTERPRETATION,
+    Interpretation,
+    Deb822ParsedTokenList,
 )
-from debputy.lsp.vendoring._deb822_repro.tokens import (
-    Deb822Token,
-)
-from debputy.util import _info
 
 try:
     from debputy.lsp.vendoring._deb822_repro.locatable import (
@@ -100,134 +109,192 @@ _LANGUAGE_IDS = [
     # vim's name
     "debcontrol",
 ]
+
+
+@dataclasses.dataclass(slots=True, frozen=True)
+class SubstvarMetadata:
+    name: str
+    defined_by: str
+    dh_sequence: Optional[str]
+    source: Optional[str]
+    description: str
+
+    def render_metadata_fields(self) -> str:
+        def_by = f"Defined by: {self.defined_by}"
+        dh_seq = (
+            f"DH Sequence: {self.dh_sequence}" if self.dh_sequence is not None else None
+        )
+        source = f"Source: {self.source}" if self.source is not None else None
+        return "\n".join(filter(None, (def_by, dh_seq, source)))
+
+
+def relationship_substvar_for_field(substvar: str) -> Optional[str]:
+    relationship_fields = all_package_relationship_fields()
+    try:
+        col_idx = substvar.rindex(":")
+    except ValueError:
+        return None
+    return relationship_fields.get(substvar[col_idx + 1 : -1].lower())
+
+
+def _substvars_metadata(*args: SubstvarMetadata) -> Mapping[str, SubstvarMetadata]:
+    r = {s.name: s for s in args}
+    assert len(r) == len(args)
+    return r
+
+
 _SUBSTVAR_RE = re.compile(r"[$][{][a-zA-Z0-9][a-zA-Z0-9-:]*[}]")
-_SUBSTVARS_DOC = {
-    "${}": textwrap.dedent(
-        """\
-        This is a substvar for a literal `$`. This form will never recurse
-        into another substvar. As an example, `${}{binary:Version}` will result
-        literal `${binary:Version}` (which will not be replaced).
-
-        Defined by: `dpkg-gencontrol`
-        DH Sequence: <default>
-        Source: <https://manpages.debian.org/deb-substvars.5>
-    """
+_SUBSTVARS_DOC = _substvars_metadata(
+    SubstvarMetadata(
+        "${}",
+        "`dpkg-gencontrol`",
+        "(default)",
+        "<https://manpages.debian.org/deb-substvars.5>",
+        textwrap.dedent(
+            """\
+            This is a substvar for a literal `$`. This form will never recurse
+            into another substvar. As an example, `${}{binary:Version}` will result
+            literal `${binary:Version}` (which will not be replaced).
+        """
+        ),
     ),
-    "${binary:Version}": textwrap.dedent(
-        """\
-        The version of the current binary package including binNMU version.
+    SubstvarMetadata(
+        "${binary:Version}",
+        "`dpkg-gencontrol`",
+        "(default)",
+        "<https://manpages.debian.org/deb-substvars.5>",
+        textwrap.dedent(
+            """\
+            The version of the current binary package including binNMU version.
 
-        Often used with `Depends: dep (= ${binary:Version})` relations
-        where:
+            Often used with `Depends: dep (= ${binary:Version})` relations
+            where:
 
-         * The `dep` package is from the same source (listed in the same
-           `debian/control` file)
-         * The current package and `dep` are both `arch:any` (or both `arch:all`)
-           packages.
-
-        Defined by: `dpkg-gencontrol`
-        DH Sequence: <default>
-        Source: <https://manpages.debian.org/deb-substvars.5>
+             * The `dep` package is from the same source (listed in the same
+               `debian/control` file)
+             * The current package and `dep` are both `arch:any` (or both `arch:all`)
+               packages.
     """
+        ),
     ),
-    "${source:Version}": textwrap.dedent(
-        """\
-        The version of the current source package excluding binNMU version.
+    SubstvarMetadata(
+        "${source:Version}",
+        "`dpkg-gencontrol`",
+        "(default)",
+        "<https://manpages.debian.org/deb-substvars.5>",
+        textwrap.dedent(
+            """\
+            The version of the current source package excluding binNMU version.
 
-        Often used with `Depends: dep (= ${source:Version})` relations
-        where:
+            Often used with `Depends: dep (= ${source:Version})` relations
+            where:
 
-         * The `dep` package is from the same source (listed in the same
-           `debian/control` file)
-         * The `dep` is `arch:all`.
-
-        Defined by: `dpkg-gencontrol`
-        DH Sequence: <default>
-        Source: <https://manpages.debian.org/deb-substvars.5>
+             * The `dep` package is from the same source (listed in the same
+               `debian/control` file)
+             * The `dep` is `arch:all`.
     """
+        ),
     ),
-    "${misc:Depends}": textwrap.dedent(
-        """\
-        Some debhelper commands may make the generated package need to depend on some other packages.
-        For example, if you use `dh_installdebconf(1)`, your package will generally need to depend on
-        debconf. Or if you use `dh_installxfonts(1)`, your package will generally need to depend on a
-        particular version of xutils. Keeping track of these miscellaneous dependencies can be
-        annoying since they are dependent on how debhelper does things, so debhelper offers a way to
-        automate it.
+    SubstvarMetadata(
+        "${misc:Depends}",
+        "`debhelper`",
+        "(default)",
+        "<https://manpages.debian.org/debhelper.7>",
+        textwrap.dedent(
+            """\
+            Some debhelper commands may make the generated package need to depend on some other packages.
+            For example, if you use `dh_installdebconf(1)`, your package will generally need to depend on
+            debconf. Or if you use `dh_installxfonts(1)`, your package will generally need to depend on a
+            particular version of xutils. Keeping track of these miscellaneous dependencies can be
+            annoying since they are dependent on how debhelper does things, so debhelper offers a way to
+            automate it.
 
-        All commands of this type, besides documenting what dependencies may be needed on their man
-        pages, will automatically generate a substvar called ${misc:Depends}. If you put that token
-        into your `debian/control` file, it will be expanded to the dependencies debhelper figures
-        you need.
+            All commands of this type, besides documenting what dependencies may be needed on their man
+            pages, will automatically generate a substvar called ${misc:Depends}. If you put that token
+            into your `debian/control` file, it will be expanded to the dependencies debhelper figures
+            you need.
 
-        This is entirely independent of the standard `${shlibs:Depends}` generated by `dh_makeshlibs(1)`,
-        and the `${perl:Depends}` generated by `dh_perl(1)`.
-
-        Defined by: `debhelper`
-        DH Sequence: <default>
-        Source: <https://manpages.debian.org/debhelper.7>
+            This is entirely independent of the standard `${shlibs:Depends}` generated by `dh_makeshlibs(1)`,
+            and the `${perl:Depends}` generated by `dh_perl(1)`.
     """
+        ),
     ),
-    "${misc:Pre-Depends}": textwrap.dedent(
-        """\
-        This is the moral equivalent to `${misc:Depends}` but for `Pre-Depends`.
-
-        Defined by: `debhelper`
-        DH Sequence: <default>
+    SubstvarMetadata(
+        "${misc:Pre-Depends}",
+        "`debhelper`",
+        "(default)",
+        None,
+        textwrap.dedent(
+            """\
+            This is the moral equivalent to `${misc:Depends}` but for `Pre-Depends`.
     """
+        ),
     ),
-    "${perl:Depends}": textwrap.dedent(
-        """\
-        The dependency on perl as determined by `dh_perl`. Note this only covers the relationship
-        with the Perl interpreter and not perl modules.
+    SubstvarMetadata(
+        "${perl:Depends}",
+        "`dh_perl`",
+        "(default)",
+        "<https://manpages.debian.org/dh_perl.1>",
+        textwrap.dedent(
+            """\
+            The dependency on perl as determined by `dh_perl`. Note this only covers the relationship
+            with the Perl interpreter and not perl modules.
 
-        Defined by: `dh_perl`
-        DH Sequence: <default>
-        Source: <https://manpages.debian.org/dh_perl.1>
     """
+        ),
     ),
-    "${gir:Depends}": textwrap.dedent(
-        """\
-        Dependencies related to GObject introspection data.
+    SubstvarMetadata(
+        "${gir:Depends}",
+        "`dh_girepository`",
+        "gir",
+        "<https://manpages.debian.org/dh_girepository.1>",
+        textwrap.dedent(
+            """\
+            Dependencies related to GObject introspection data.
+    """
+        ),
+    ),
+    SubstvarMetadata(
+        "${shlibs:Depends}",
+        "`dpkg-shlibdeps` (often via `dh_shlibdeps`)",
+        "(default)",
+        "<https://manpages.debian.org/dpkg-shlibdeps.1>",
+        textwrap.dedent(
+            """\
+            Dependencies related to ELF dependencies.
+    """
+        ),
+    ),
+    SubstvarMetadata(
+        "${shlibs:Pre-Depends}",
+        "`dpkg-shlibdeps` (often via `dh_shlibdeps`)",
+        "(default)",
+        "<https://manpages.debian.org/dpkg-shlibdeps.1>",
+        textwrap.dedent(
+            """\
+            Dependencies related to ELF dependencies. The `Pre-Depends`
+            version is often only seen in `Essential: yes` packages
+            or packages that manually request the `Pre-Depends`
+            relation via `dpkg-shlibdeps`.
 
-        Defined by: `dh_girepository`
-        DH Sequence: `gir`
-        Source: <https://manpages.debian.org/dh_girepository.1>
+            Note: This substvar only appears in `debhelper-compat (= 14)`, or
+            with use of `debputy` (at an integration level, where `debputy`
+            runs `dpkg-shlibdeps`), or when passing relevant options to
+            `dpkg-shlibdeps`  (often via `dh_shlibdeps`) such as `-dPre-Depends`.
     """
+        ),
     ),
-    "${shlibs:Depends}": textwrap.dedent(
-        """\
-        Dependencies related to ELF dependencies.
-
-        Defined by: `dpkg-shlibdeps` (often via `dh_shlibdeps`)
-        DH Sequence: <default>
-        Source: <https://manpages.debian.org/dpkg-shlibdeps.1>
-    """
-    ),
-    "${shlibs:Pre-Depends}": textwrap.dedent(
-        """\
-        Dependencies related to ELF dependencies. The `Pre-Depends`
-        version is often only seen in `Essential: yes` packages
-        or packages that manually request the `Pre-Depends`
-        relation via `dpkg-shlibdeps`.
-
-        Defined by: `dpkg-shlibdeps` (often via `dh_shlibdeps`)
-        DH Sequence: <default>
-        Source: <https://manpages.debian.org/dpkg-shlibdeps.1>
-    """
-    ),
-}
+)
 
 _DCTRL_FILE_METADATA = DctrlFileMetadata()
 
 
 lsp_standard_handler(_LANGUAGE_IDS, TEXT_DOCUMENT_CODE_ACTION)
-lsp_standard_handler(_LANGUAGE_IDS, TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL)
 
 
 @lsp_hover(_LANGUAGE_IDS)
 def _debian_control_hover(
-    ls: "LanguageServer",
+    ls: "DebputyLanguageServer",
     params: HoverParams,
 ) -> Optional[Hover]:
     return deb822_hover(ls, params, _DCTRL_FILE_METADATA, custom_handler=_custom_hover)
@@ -248,26 +315,40 @@ def _custom_hover(
     line_no = server_position.line
     line = lines[line_no]
     substvar_search_ref = server_position.character
-    if line[substvar_search_ref] in ("$", "{"):
-        substvar_search_ref += 2
     substvar = ""
     try:
+        if line and line[substvar_search_ref] in ("$", "{"):
+            substvar_search_ref += 2
         substvar_start = line.rindex("${", 0, substvar_search_ref)
         substvar_end = line.index("}", substvar_start)
         if server_position.character <= substvar_end:
-            _info(
-                f"Range {substvar_start} <= {server_position.character} <= {substvar_end}"
-            )
             substvar = line[substvar_start : substvar_end + 1]
-    except ValueError:
+    except (ValueError, IndexError):
         pass
 
     if substvar == "${}" or _SUBSTVAR_RE.fullmatch(substvar):
-        doc = _SUBSTVARS_DOC.get(substvar)
+        substvar_md = _SUBSTVARS_DOC.get(substvar)
 
-        if doc is None:
-            doc = "No documentation for {substvar}."
-        return f"# Substvar `{substvar}`\n\n{doc}"
+        computed_doc = ""
+        for_field = relationship_substvar_for_field(substvar)
+        if for_field:
+            # Leading empty line is intentional!
+            computed_doc = textwrap.dedent(
+                f"""
+                This substvar is a relationship substvar for the field {for_field}.
+                Relationship substvars are automatically added in the field they
+                are named after in `debhelper-compat (= 14)` or later, or with
+                `debputy` (any integration mode after 0.1.21).
+            """
+            )
+
+        if substvar_md is None:
+            doc = f"No documentation for {substvar}.\n"
+            md_fields = ""
+        else:
+            doc = substvar_md.description
+            md_fields = "\n" + substvar_md.render_metadata_fields()
+        return f"# Substvar `{substvar}`\n\n{doc}{computed_doc}{md_fields}"
 
     if known_field is None or known_field.name != "Description":
         return None
@@ -279,14 +360,41 @@ def _custom_hover(
         return None
 
     content = line[col_idx + 1 :].strip()
+
     # Synopsis
     return textwrap.dedent(
         f"""\
         # Package synopsis
 
-        The synopsis is a single line "noun phrase" description of the package.
-        It is typically used in search results and other cases where a
-        user-interface has limited space for text.
+        The synopsis functions as a phrase describing the package, not a
+        complete sentence, so sentential punctuation is inappropriate: it
+        does not need extra capital letters or a final period (full stop).
+        It should also omit any initial indefinite or definite article
+        - "a", "an", or "the". Thus for instance:
+
+        ```
+        Package: libeg0
+        Description: exemplification support library
+        ```
+
+        Technically this is a noun phrase minus articles, as opposed to a
+        verb phrase. A good heuristic is that it should be possible to
+        substitute the package name and synopsis into this formula:
+
+        ```
+        # Generic
+        The package provides {{a,an,the,some}} synopsis.
+
+        # The current package for comparison
+        The package provides {{a,an,the,some}} {content}.
+        ```
+
+        Other advice for writing synopsis:
+         * Avoid using the package name. Any software would display the
+           package name already and it generally does not help the user
+           understand what they are looking at.
+         * In many situations, the user will only see the package name
+           and its synopsis. The synopsis must standalone.
 
         **Example renderings in various terminal UIs**:
         ```
@@ -298,19 +406,34 @@ def _custom_hover(
         package - {content}
         ```
 
-        Advice for writing synopsis:
-         * Avoid using the package name. Any software would display the
-           package name already and it generally does not help the user
-           understand what they are looking at.
-         * In many situations, the user will only see the package name
-           and its synopsis. The synopsis must standalone.
-         * When writing the synopsis, it is often a good idea to write
-           it such that it fits into the sentence like "This package
-           provides [a|an|the] ..." (see below).
+        ## Reference example
+
+        An reference example for comparison: The Sphinx package
+        (python3-sphinx/7.2.6-6) had the following synopsis:
 
-        **Phrasing test**:
         ```
-        This package provides [a|an|the] {content}.
+        Description: documentation generator for Python projects
+        ```
+
+        In the test sentence, it would read as:
+
+        ```
+        The python3-sphinx package provides a documentation generator for Python projects.
+        ```
+
+        **Side-by-side comparison in the terminal UIs**:
+        ```
+        # apt search TERM
+        python3-sphinx/stable,now 7.2.6-6 all:
+           documentation generator for Python projects
+
+        package/stable,now 1.0-1 all:
+           {content}
+
+
+        # apt-get search TERM
+        package - {content}
+        python3-sphinx - documentation generator for Python projects
         ```
     """
     )
@@ -318,7 +441,7 @@ def _custom_hover(
 
 @lsp_completer(_LANGUAGE_IDS)
 def _debian_control_completions(
-    ls: "LanguageServer",
+    ls: "DebputyLanguageServer",
     params: CompletionParams,
 ) -> Optional[Union[CompletionList, Sequence[CompletionItem]]]:
     return deb822_completer(ls, params, _DCTRL_FILE_METADATA)
@@ -326,67 +449,58 @@ def _debian_control_completions(
 
 @lsp_folding_ranges(_LANGUAGE_IDS)
 def _debian_control_folding_ranges(
-    ls: "LanguageServer",
+    ls: "DebputyLanguageServer",
     params: FoldingRangeParams,
 ) -> Optional[Sequence[FoldingRange]]:
     return deb822_folding_ranges(ls, params, _DCTRL_FILE_METADATA)
 
 
-def _deb822_token_iter(
-    tokens: Iterable[Deb822Token],
-) -> Iterator[Tuple[Deb822Token, int, int, int, int, int]]:
-    line_no = 0
-    line_offset = 0
-
-    for token in tokens:
-        start_line = line_no
-        start_line_offset = line_offset
-
-        newlines = token.text.count("\n")
-        line_no += newlines
-        text_len = len(token.text)
-        if newlines:
-            if token.text.endswith("\n"):
-                line_offset = 0
-            else:
-                # -2, one to remove the "\n" and one to get 0-offset
-                line_offset = text_len - token.text.rindex("\n") - 2
-        else:
-            line_offset += text_len
-
-        yield token, start_line, start_line_offset, line_no, line_offset
-
-
 def _paragraph_representation_field(
     paragraph: Deb822ParagraphElement,
 ) -> Deb822KeyValuePairElement:
     return next(iter(paragraph.iter_parts_of_type(Deb822KeyValuePairElement)))
 
 
-def _extract_first_value_and_position(
-    kvpair: Deb822KeyValuePairElement,
-    stanza_pos: "TEPosition",
-    position_codec: "LintCapablePositionCodec",
-    lines: List[str],
-) -> Tuple[Optional[str], Optional[Range]]:
-    kvpair_pos = kvpair.position_in_parent().relative_to(stanza_pos)
-    value_element_pos = kvpair.value_element.position_in_parent().relative_to(
-        kvpair_pos
-    )
-    for value_ref in kvpair.interpret_as(
-        LIST_SPACE_SEPARATED_INTERPRETATION
-    ).iter_value_references():
-        v = value_ref.value
-        section_value_loc = value_ref.locatable
-        value_range_te = section_value_loc.range_in_parent().relative_to(
-            value_element_pos
+def _source_package_checks(
+    stanza: Deb822ParagraphElement,
+    stanza_position: "TEPosition",
+    lint_state: LintState,
+    diagnostics: List[Diagnostic],
+) -> None:
+    vcs_fields = {}
+    for kvpair in stanza.iter_parts_of_type(Deb822KeyValuePairElement):
+        name = normalize_dctrl_field_name(kvpair.field_name.lower())
+        if (
+            not name.startswith("vcs-")
+            or name == "vcs-browser"
+            or name not in SOURCE_FIELDS
+        ):
+            continue
+        vcs_fields[name] = kvpair
+
+    if len(vcs_fields) < 2:
+        return
+    for kvpair in vcs_fields.values():
+        kvpair_range_server_units = te_range_to_lsp(
+            kvpair.range_in_parent().relative_to(stanza_position)
         )
-        section_range_server_units = te_range_to_lsp(value_range_te)
-        section_range = position_codec.range_to_client_units(
-            lines, section_range_server_units
+        diagnostics.append(
+            Diagnostic(
+                lint_state.position_codec.range_to_client_units(
+                    lint_state.lines, kvpair_range_server_units
+                ),
+                f'Multiple Version Control fields defined ("{kvpair.field_name}")',
+                severity=DiagnosticSeverity.Warning,
+                source="debputy",
+                data=DiagnosticData(
+                    quickfixes=[
+                        propose_remove_range_quick_fix(
+                            proposed_title=f'Remove "{kvpair.field_name}"'
+                        )
+                    ]
+                ),
+            )
         )
-        return v, section_range
-    return None, None
 
 
 def _binary_package_checks(
@@ -394,8 +508,7 @@ def _binary_package_checks(
     stanza_position: "TEPosition",
     source_stanza: Deb822ParagraphElement,
     representation_field_range: Range,
-    position_codec: "LintCapablePositionCodec",
-    lines: List[str],
+    lint_state: LintState,
     diagnostics: List[Diagnostic],
 ) -> None:
     package_name = stanza.get("Package", "")
@@ -403,11 +516,10 @@ def _binary_package_checks(
     section_kvpair = stanza.get_kvpair_element("Section", use_get=True)
     section: Optional[str] = None
     if section_kvpair is not None:
-        section, section_range = _extract_first_value_and_position(
+        section, section_range = extract_first_value_and_position(
             section_kvpair,
             stanza_position,
-            position_codec,
-            lines,
+            lint_state,
         )
     else:
         section_range = representation_field_range
@@ -425,11 +537,10 @@ def _binary_package_checks(
             )
             package_type_range = None
             if package_type_kvpair is not None:
-                _, package_type_range = _extract_first_value_and_position(
+                _, package_type_range = extract_first_value_and_position(
                     package_type_kvpair,
                     stanza_position,
-                    position_codec,
-                    lines,
+                    lint_state,
                 )
             if package_type_range is None:
                 package_type_range = representation_field_range
@@ -441,26 +552,36 @@ def _binary_package_checks(
                     source="debputy",
                 )
             )
-        if effective_section != "debian-installer":
-            quickfix_data = None
-            if section is not None:
-                quickfix_data = [
-                    propose_correct_text_quick_fix(
-                        f"{component_prefix}debian-installer"
-                    )
-                ]
-            diagnostics.append(
-                Diagnostic(
-                    section_range,
-                    f'The Section should be "{component_prefix}debian-installer" for udebs',
-                    severity=DiagnosticSeverity.Warning,
-                    source="debputy",
-                    data=quickfix_data,
+        guessed_section = "debian-installer"
+        section_diagnostic_rationale = " since it is an udeb"
+    else:
+        guessed_section = package_name_to_section(package_name)
+        section_diagnostic_rationale = " based on the package name"
+    if guessed_section is not None and guessed_section != effective_section:
+        if section is not None:
+            quickfix_data = [
+                propose_correct_text_quick_fix(f"{component_prefix}{guessed_section}")
+            ]
+        else:
+            quickfix_data = [
+                propose_insert_text_on_line_after_diagnostic_quick_fix(
+                    f"Section: {component_prefix}{guessed_section}\n"
                 )
+            ]
+        assert section_range is not None  # mypy hint
+        diagnostics.append(
+            Diagnostic(
+                section_range,
+                f'The Section should be "{component_prefix}{guessed_section}"{section_diagnostic_rationale}',
+                severity=DiagnosticSeverity.Warning,
+                source="debputy",
+                data=DiagnosticData(quickfixes=quickfix_data),
             )
+        )
 
 
 def _diagnostics_for_paragraph(
+    deb822_file: Deb822FileElement,
     stanza: Deb822ParagraphElement,
     stanza_position: "TEPosition",
     source_stanza: Deb822ParagraphElement,
@@ -468,37 +589,29 @@ def _diagnostics_for_paragraph(
     other_known_fields: Mapping[str, DctrlKnownField],
     is_binary_paragraph: bool,
     doc_reference: str,
-    position_codec: "LintCapablePositionCodec",
-    lines: List[str],
+    lint_state: LintState,
     diagnostics: List[Diagnostic],
 ) -> None:
     representation_field = _paragraph_representation_field(stanza)
-    representation_field_pos = representation_field.position_in_parent().relative_to(
+    representation_field_range = representation_field.range_in_parent().relative_to(
         stanza_position
     )
-    representation_field_range_server_units = te_range_to_lsp(
-        TERange.from_position_and_size(
-            representation_field_pos, representation_field.size()
-        )
-    )
-    representation_field_range = position_codec.range_to_client_units(
-        lines,
-        representation_field_range_server_units,
+    representation_field_range = lint_state.position_codec.range_to_client_units(
+        lint_state.lines,
+        te_range_to_lsp(representation_field_range),
     )
     for known_field in known_fields.values():
-        missing_field_severity = known_field.missing_field_severity
-        if missing_field_severity is None or known_field.name in stanza:
+        if known_field.name in stanza:
             continue
 
-        if known_field.inherits_from_source and known_field.name in source_stanza:
-            continue
-
-        diagnostics.append(
-            Diagnostic(
+        diagnostics.extend(
+            known_field.field_omitted_diagnostics(
+                deb822_file,
                 representation_field_range,
-                f"Stanza is missing field {known_field.name}",
-                severity=missing_field_severity,
-                source="debputy",
+                stanza,
+                stanza_position,
+                source_stanza,
+                lint_state,
             )
         )
 
@@ -508,12 +621,18 @@ def _diagnostics_for_paragraph(
             stanza_position,
             source_stanza,
             representation_field_range,
-            position_codec,
-            lines,
+            lint_state,
+            diagnostics,
+        )
+    else:
+        _source_package_checks(
+            stanza,
+            stanza_position,
+            lint_state,
             diagnostics,
         )
 
-    seen_fields = {}
+    seen_fields: Dict[str, Tuple[str, str, Range, List[Range]]] = {}
 
     for kvpair in stanza.iter_parts_of_type(Deb822KeyValuePairElement):
         field_name_token = kvpair.field_token
@@ -522,11 +641,14 @@ def _diagnostics_for_paragraph(
         normalized_field_name_lc = normalize_dctrl_field_name(field_name_lc)
         known_field = known_fields.get(normalized_field_name_lc)
         field_value = stanza[field_name]
-        field_range_te = kvpair.range_in_parent().relative_to(stanza_position)
+        kvpair_position = kvpair.position_in_parent().relative_to(stanza_position)
+        field_range_te = kvpair.field_token.range_in_parent().relative_to(
+            kvpair_position
+        )
         field_position_te = field_range_te.start_pos
         field_range_server_units = te_range_to_lsp(field_range_te)
-        field_range = position_codec.range_to_client_units(
-            lines,
+        field_range = lint_state.position_codec.range_to_client_units(
+            lint_state.lines,
             field_range_server_units,
         )
         field_name_typo_detected = False
@@ -551,8 +673,8 @@ def _diagnostics_for_paragraph(
                         field_position_te, kvpair.field_token.size()
                     )
                 )
-                field_range = position_codec.range_to_client_units(
-                    lines,
+                field_range = lint_state.position_codec.range_to_client_units(
+                    lint_state.lines,
                     token_range_server_units,
                 )
                 field_name_typo_detected = True
@@ -562,10 +684,12 @@ def _diagnostics_for_paragraph(
                         f'The "{field_name}" looks like a typo of "{known_field.name}".',
                         severity=DiagnosticSeverity.Warning,
                         source="debputy",
-                        data=[
-                            propose_correct_text_quick_fix(known_fields[m].name)
-                            for m in candidates
-                        ],
+                        data=DiagnosticData(
+                            quickfixes=[
+                                propose_correct_text_quick_fix(known_fields[m].name)
+                                for m in candidates
+                            ]
+                        ),
                     )
                 )
         if known_field is None:
@@ -595,11 +719,12 @@ def _diagnostics_for_paragraph(
             continue
         diagnostics.extend(
             known_field.field_diagnostics(
+                deb822_file,
                 kvpair,
                 stanza,
                 stanza_position,
-                position_codec,
-                lines,
+                kvpair_position,
+                lint_state,
                 field_name_typo_reported=field_name_typo_detected,
             )
         )
@@ -621,15 +746,15 @@ def _diagnostics_for_paragraph(
                     )
                     if pos:
                         word_pos_te = TEPosition(0, pos).relative_to(word_pos_te)
-                    word_range = TERange(
+                    word_range_te = TERange(
                         START_POSITION,
                         TEPosition(0, endpos - pos),
                     )
                     word_range_server_units = te_range_to_lsp(
-                        TERange.from_position_and_size(word_pos_te, word_range)
+                        TERange.from_position_and_size(word_pos_te, word_range_te)
                     )
-                    word_range = position_codec.range_to_client_units(
-                        lines,
+                    word_range = lint_state.position_codec.range_to_client_units(
+                        lint_state.lines,
                         word_range_server_units,
                     )
                     diagnostics.append(
@@ -638,9 +763,13 @@ def _diagnostics_for_paragraph(
                             f'Spelling "{word}"',
                             severity=DiagnosticSeverity.Hint,
                             source="debputy",
-                            data=[
-                                propose_correct_text_quick_fix(c) for c in corrections
-                            ],
+                            data=DiagnosticData(
+                                lint_severity="spelling",
+                                quickfixes=[
+                                    propose_correct_text_quick_fix(c)
+                                    for c in corrections
+                                ],
+                            ),
                         )
                     )
         source_value = source_stanza.get(field_name)
@@ -666,7 +795,7 @@ def _diagnostics_for_paragraph(
                     f"The field {field_name} duplicates the value from the Source stanza.",
                     severity=DiagnosticSeverity.Information,
                     source="debputy",
-                    data=fix_data,
+                    data=DiagnosticData(quickfixes=fix_data),
                 )
             )
     for (
@@ -718,7 +847,7 @@ def _scan_for_syntax_errors_and_token_level_diagnostics(
         start_offset,
         end_line,
         end_offset,
-    ) in _deb822_token_iter(deb822_file.iter_tokens()):
+    ) in deb822_token_iter(deb822_file.iter_tokens()):
         if token.is_error:
             first_error = min(first_error, start_line)
             start_pos = Position(
@@ -741,17 +870,17 @@ def _scan_for_syntax_errors_and_token_level_diagnostics(
                 )
             )
         elif token.is_comment:
-            for word, pos, end_pos in spell_checker.iter_words(token.text):
+            for word, col_pos, end_col_pos in spell_checker.iter_words(token.text):
                 corrections = spell_checker.provide_corrections_for(word)
                 if not corrections:
                     continue
                 start_pos = Position(
                     start_line,
-                    pos,
+                    col_pos,
                 )
                 end_pos = Position(
                     start_line,
-                    end_pos,
+                    end_col_pos,
                 )
                 word_range = position_codec.range_to_client_units(
                     lines, Range(start_pos, end_pos)
@@ -762,7 +891,12 @@ def _scan_for_syntax_errors_and_token_level_diagnostics(
                         f'Spelling "{word}"',
                         severity=DiagnosticSeverity.Hint,
                         source="debputy",
-                        data=[propose_correct_text_quick_fix(c) for c in corrections],
+                        data=DiagnosticData(
+                            lint_severity="spelling",
+                            quickfixes=[
+                                propose_correct_text_quick_fix(c) for c in corrections
+                            ],
+                        ),
                     )
                 )
     return first_error
@@ -776,11 +910,7 @@ def _lint_debian_control(
     position_codec = lint_state.position_codec
     doc_reference = lint_state.doc_uri
     diagnostics = []
-    deb822_file = parse_deb822_file(
-        lines,
-        accept_files_with_duplicated_fields=True,
-        accept_files_with_error_tokens=True,
-    )
+    deb822_file = lint_state.parsed_deb822_file_content
 
     first_error = _scan_for_syntax_errors_and_token_level_diagnostics(
         deb822_file,
@@ -804,6 +934,7 @@ def _lint_debian_control(
             known_fields = SOURCE_FIELDS
             other_known_fields = BINARY_FIELDS
         _diagnostics_for_paragraph(
+            deb822_file,
             paragraph,
             paragraph_pos,
             source_paragraph,
@@ -811,17 +942,42 @@ def _lint_debian_control(
             other_known_fields,
             is_binary_paragraph,
             doc_reference,
-            position_codec,
-            lines,
+            lint_state,
             diagnostics,
         )
 
     return diagnostics
 
 
+@lsp_will_save_wait_until(_LANGUAGE_IDS)
+def _debian_control_on_save_formatting(
+    ls: "DebputyLanguageServer",
+    params: WillSaveTextDocumentParams,
+) -> Optional[Sequence[TextEdit]]:
+    doc = ls.workspace.get_text_document(params.text_document.uri)
+    lint_state = ls.lint_state(doc)
+    return _reformat_debian_control(lint_state)
+
+
+def _reformat_debian_control(
+    lint_state: LintState,
+) -> Optional[Sequence[TextEdit]]:
+    return deb822_format_file(lint_state, _DCTRL_FILE_METADATA)
+
+
+@lsp_format_document(_LANGUAGE_IDS)
+def _debian_control_on_save_formatting(
+    ls: "DebputyLanguageServer",
+    params: DocumentFormattingParams,
+) -> Optional[Sequence[TextEdit]]:
+    doc = ls.workspace.get_text_document(params.text_document.uri)
+    lint_state = ls.lint_state(doc)
+    return deb822_format_file(lint_state, _DCTRL_FILE_METADATA)
+
+
 @lsp_semantic_tokens_full(_LANGUAGE_IDS)
-def _semantic_tokens_full(
-    ls: "LanguageServer",
+def _debian_control_semantic_tokens_full(
+    ls: "DebputyLanguageServer",
     request: SemanticTokensParams,
 ) -> Optional[SemanticTokens]:
     return deb822_semantic_tokens_full(
