@@ -3,11 +3,10 @@ from typing import (
     Union,
     Sequence,
     Tuple,
-    Iterator,
     Optional,
-    Iterable,
     Mapping,
     List,
+    Dict,
 )
 
 from lsprotocol.types import (
@@ -18,7 +17,6 @@ from lsprotocol.types import (
     CompletionItem,
     CompletionList,
     CompletionParams,
-    TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL,
     DiagnosticRelatedInformation,
     Location,
     HoverParams,
@@ -28,9 +26,14 @@ from lsprotocol.types import (
     SemanticTokensParams,
     FoldingRangeParams,
     FoldingRange,
+    WillSaveTextDocumentParams,
+    TextEdit,
+    DocumentFormattingParams,
 )
 
 from debputy.linting.lint_util import LintState
+from debputy.lsp.debputy_ls import DebputyLanguageServer
+from debputy.lsp.diagnostics import DiagnosticData
 from debputy.lsp.lsp_debian_control_reference_data import (
     _DEP5_HEADER_FIELDS,
     _DEP5_FILES_FIELDS,
@@ -45,12 +48,16 @@ from debputy.lsp.lsp_features import (
     lsp_standard_handler,
     lsp_folding_ranges,
     lsp_semantic_tokens_full,
+    lsp_will_save_wait_until,
+    lsp_format_document,
 )
 from debputy.lsp.lsp_generic_deb822 import (
     deb822_completer,
     deb822_hover,
     deb822_folding_ranges,
     deb822_semantic_tokens_full,
+    deb822_token_iter,
+    deb822_format_file,
 )
 from debputy.lsp.quickfixes import (
     propose_correct_text_quick_fix,
@@ -63,7 +70,6 @@ from debputy.lsp.text_util import (
     te_range_to_lsp,
 )
 from debputy.lsp.vendoring._deb822_repro import (
-    parse_deb822_file,
     Deb822FileElement,
     Deb822ParagraphElement,
 )
@@ -71,9 +77,6 @@ from debputy.lsp.vendoring._deb822_repro.parsing import (
     Deb822KeyValuePairElement,
     LIST_SPACE_SEPARATED_INTERPRETATION,
 )
-from debputy.lsp.vendoring._deb822_repro.tokens import (
-    Deb822Token,
-)
 
 try:
     from debputy.lsp.vendoring._deb822_repro.locatable import (
@@ -100,12 +103,11 @@ _LANGUAGE_IDS = [
 _DEP5_FILE_METADATA = Dep5FileMetadata()
 
 lsp_standard_handler(_LANGUAGE_IDS, TEXT_DOCUMENT_CODE_ACTION)
-lsp_standard_handler(_LANGUAGE_IDS, TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL)
 
 
 @lsp_hover(_LANGUAGE_IDS)
 def _debian_copyright_hover(
-    ls: "LanguageServer",
+    ls: "DebputyLanguageServer",
     params: HoverParams,
 ) -> Optional[Hover]:
     return deb822_hover(ls, params, _DEP5_FILE_METADATA)
@@ -113,7 +115,7 @@ def _debian_copyright_hover(
 
 @lsp_completer(_LANGUAGE_IDS)
 def _debian_copyright_completions(
-    ls: "LanguageServer",
+    ls: "DebputyLanguageServer",
     params: CompletionParams,
 ) -> Optional[Union[CompletionList, Sequence[CompletionItem]]]:
     return deb822_completer(ls, params, _DEP5_FILE_METADATA)
@@ -121,37 +123,12 @@ def _debian_copyright_completions(
 
 @lsp_folding_ranges(_LANGUAGE_IDS)
 def _debian_copyright_folding_ranges(
-    ls: "LanguageServer",
+    ls: "DebputyLanguageServer",
     params: FoldingRangeParams,
 ) -> Optional[Sequence[FoldingRange]]:
     return deb822_folding_ranges(ls, params, _DEP5_FILE_METADATA)
 
 
-def _deb822_token_iter(
-    tokens: Iterable[Deb822Token],
-) -> Iterator[Tuple[Deb822Token, int, int, int, int, int]]:
-    line_no = 0
-    line_offset = 0
-
-    for token in tokens:
-        start_line = line_no
-        start_line_offset = line_offset
-
-        newlines = token.text.count("\n")
-        line_no += newlines
-        text_len = len(token.text)
-        if newlines:
-            if token.text.endswith("\n"):
-                line_offset = 0
-            else:
-                # -2, one to remove the "\n" and one to get 0-offset
-                line_offset = text_len - token.text.rindex("\n") - 2
-        else:
-            line_offset += text_len
-
-        yield token, start_line, start_line_offset, line_no, line_offset
-
-
 def _paragraph_representation_field(
     paragraph: Deb822ParagraphElement,
 ) -> Deb822KeyValuePairElement:
@@ -159,14 +136,14 @@ def _paragraph_representation_field(
 
 
 def _diagnostics_for_paragraph(
+    deb822_file: Deb822FileElement,
     stanza: Deb822ParagraphElement,
     stanza_position: "TEPosition",
     known_fields: Mapping[str, Deb822KnownField],
     other_known_fields: Mapping[str, Deb822KnownField],
     is_files_or_license_paragraph: bool,
     doc_reference: str,
-    position_codec: "LintCapablePositionCodec",
-    lines: List[str],
+    lint_state: LintState,
     diagnostics: List[Diagnostic],
 ) -> None:
     representation_field = _paragraph_representation_field(stanza)
@@ -178,25 +155,26 @@ def _diagnostics_for_paragraph(
             representation_field_pos, representation_field.size()
         )
     )
-    representation_field_range = position_codec.range_to_client_units(
-        lines,
+    representation_field_range = lint_state.position_codec.range_to_client_units(
+        lint_state.lines,
         representation_field_range_server_units,
     )
     for known_field in known_fields.values():
-        missing_field_severity = known_field.missing_field_severity
-        if missing_field_severity is None or known_field.name in stanza:
+        if known_field.name in stanza:
             continue
 
-        diagnostics.append(
-            Diagnostic(
+        diagnostics.extend(
+            known_field.field_omitted_diagnostics(
+                deb822_file,
                 representation_field_range,
-                f"Stanza is missing field {known_field.name}",
-                severity=missing_field_severity,
-                source="debputy",
+                stanza,
+                stanza_position,
+                None,
+                lint_state,
             )
         )
 
-    seen_fields = {}
+    seen_fields: Dict[str, Tuple[str, str, Range, List[Range]]] = {}
 
     for kvpair in stanza.iter_parts_of_type(Deb822KeyValuePairElement):
         field_name_token = kvpair.field_token
@@ -205,11 +183,14 @@ def _diagnostics_for_paragraph(
         normalized_field_name_lc = normalize_dctrl_field_name(field_name_lc)
         known_field = known_fields.get(normalized_field_name_lc)
         field_value = stanza[field_name]
-        field_range_te = kvpair.range_in_parent().relative_to(stanza_position)
+        kvpair_position = kvpair.position_in_parent().relative_to(stanza_position)
+        field_range_te = kvpair.field_token.range_in_parent().relative_to(
+            kvpair_position
+        )
         field_position_te = field_range_te.start_pos
         field_range_server_units = te_range_to_lsp(field_range_te)
-        field_range = position_codec.range_to_client_units(
-            lines,
+        field_range = lint_state.position_codec.range_to_client_units(
+            lint_state.lines,
             field_range_server_units,
         )
         field_name_typo_detected = False
@@ -234,8 +215,8 @@ def _diagnostics_for_paragraph(
                         field_position_te, kvpair.field_token.size()
                     )
                 )
-                field_range = position_codec.range_to_client_units(
-                    lines,
+                field_range = lint_state.position_codec.range_to_client_units(
+                    lint_state.lines,
                     token_range_server_units,
                 )
                 field_name_typo_detected = True
@@ -245,10 +226,12 @@ def _diagnostics_for_paragraph(
                         f'The "{field_name}" looks like a typo of "{known_field.name}".',
                         severity=DiagnosticSeverity.Warning,
                         source="debputy",
-                        data=[
-                            propose_correct_text_quick_fix(known_fields[m].name)
-                            for m in candidates
-                        ],
+                        data=DiagnosticData(
+                            quickfixes=[
+                                propose_correct_text_quick_fix(known_fields[m].name)
+                                for m in candidates
+                            ]
+                        ),
                     )
                 )
         if known_field is None:
@@ -280,11 +263,12 @@ def _diagnostics_for_paragraph(
             continue
         diagnostics.extend(
             known_field.field_diagnostics(
+                deb822_file,
                 kvpair,
                 stanza,
                 stanza_position,
-                position_codec,
-                lines,
+                kvpair_position,
+                lint_state,
                 field_name_typo_reported=field_name_typo_detected,
             )
         )
@@ -306,15 +290,15 @@ def _diagnostics_for_paragraph(
                     )
                     if pos:
                         word_pos_te = TEPosition(0, pos).relative_to(word_pos_te)
-                    word_range = TERange(
+                    word_range_te = TERange(
                         START_POSITION,
                         TEPosition(0, endpos - pos),
                     )
                     word_range_server_units = te_range_to_lsp(
-                        TERange.from_position_and_size(word_pos_te, word_range)
+                        TERange.from_position_and_size(word_pos_te, word_range_te)
                     )
-                    word_range = position_codec.range_to_client_units(
-                        lines,
+                    word_range = lint_state.position_codec.range_to_client_units(
+                        lint_state.lines,
                         word_range_server_units,
                     )
                     diagnostics.append(
@@ -323,9 +307,13 @@ def _diagnostics_for_paragraph(
                             f'Spelling "{word}"',
                             severity=DiagnosticSeverity.Hint,
                             source="debputy",
-                            data=[
-                                propose_correct_text_quick_fix(c) for c in corrections
-                            ],
+                            data=DiagnosticData(
+                                lint_severity="spelling",
+                                quickfixes=[
+                                    propose_correct_text_quick_fix(c)
+                                    for c in corrections
+                                ],
+                            ),
                         )
                     )
         if known_field.warn_if_default and field_value == known_field.default_value:
@@ -387,7 +375,7 @@ def _scan_for_syntax_errors_and_token_level_diagnostics(
         start_offset,
         end_line,
         end_offset,
-    ) in _deb822_token_iter(deb822_file.iter_tokens()):
+    ) in deb822_token_iter(deb822_file.iter_tokens()):
         if token.is_error:
             first_error = min(first_error, start_line)
             start_pos = Position(
@@ -431,7 +419,12 @@ def _scan_for_syntax_errors_and_token_level_diagnostics(
                         f'Spelling "{word}"',
                         severity=DiagnosticSeverity.Hint,
                         source="debputy",
-                        data=[propose_correct_text_quick_fix(c) for c in corrections],
+                        data=DiagnosticData(
+                            lint_severity="spelling",
+                            quickfixes=[
+                                propose_correct_text_quick_fix(c) for c in corrections
+                            ],
+                        ),
                     )
                 )
     return first_error
@@ -444,12 +437,8 @@ def _lint_debian_copyright(
     lines = lint_state.lines
     position_codec = lint_state.position_codec
     doc_reference = lint_state.doc_uri
-    diagnostics = []
-    deb822_file = parse_deb822_file(
-        lines,
-        accept_files_with_duplicated_fields=True,
-        accept_files_with_error_tokens=True,
-    )
+    diagnostics: List[Diagnostic] = []
+    deb822_file = lint_state.parsed_deb822_file_content
 
     first_error = _scan_for_syntax_errors_and_token_level_diagnostics(
         deb822_file,
@@ -478,14 +467,14 @@ def _lint_debian_copyright(
         else:
             break
         _diagnostics_for_paragraph(
+            deb822_file,
             paragraph,
             paragraph_pos,
             known_fields,
             other_known_fields,
             is_files_or_license_paragraph,
             doc_reference,
-            position_codec,
-            lines,
+            lint_state,
             diagnostics,
         )
     if not is_dep5:
@@ -493,9 +482,35 @@ def _lint_debian_copyright(
     return diagnostics
 
 
+@lsp_will_save_wait_until(_LANGUAGE_IDS)
+def _debian_copyright_on_save_formatting(
+    ls: "DebputyLanguageServer",
+    params: WillSaveTextDocumentParams,
+) -> Optional[Sequence[TextEdit]]:
+    doc = ls.workspace.get_text_document(params.text_document.uri)
+    lint_state = ls.lint_state(doc)
+    return deb822_format_file(lint_state, _DEP5_FILE_METADATA)
+
+
+def _reformat_debian_copyright(
+    lint_state: LintState,
+) -> Optional[Sequence[TextEdit]]:
+    return deb822_format_file(lint_state, _DEP5_FILE_METADATA)
+
+
+@lsp_format_document(_LANGUAGE_IDS)
+def _debian_copyright_on_save_formatting(
+    ls: "DebputyLanguageServer",
+    params: DocumentFormattingParams,
+) -> Optional[Sequence[TextEdit]]:
+    doc = ls.workspace.get_text_document(params.text_document.uri)
+    lint_state = ls.lint_state(doc)
+    return deb822_format_file(lint_state, _DEP5_FILE_METADATA)
+
+
 @lsp_semantic_tokens_full(_LANGUAGE_IDS)
-def _semantic_tokens_full(
-    ls: "LanguageServer",
+def _debian_copyright_semantic_tokens_full(
+    ls: "DebputyLanguageServer",
     request: SemanticTokensParams,
 ) -> Optional[SemanticTokens]:
     return deb822_semantic_tokens_full(
