@@ -30,6 +30,7 @@ from lsprotocol.types import (
 )
 
 from debputy.linting.lint_util import LintState
+from debputy.lsp.diagnostics import DiagnosticData
 from debputy.lsp.quickfixes import propose_correct_text_quick_fix
 from debputy.manifest_parser.base_types import DebputyDispatchableType
 from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
@@ -134,7 +135,7 @@ def _lint_debian_debputy_manifest(
     path = lint_state.path
     if not is_valid_file(path):
         return None
-    diagnostics = []
+    diagnostics: List[Diagnostic] = []
     try:
         content = MANIFEST_YAML.load("".join(lines))
     except MarkedYAMLError as e:
@@ -227,7 +228,9 @@ def _unknown_key(
         message_format.format(key=key) + extra,
         DiagnosticSeverity.Error,
         source="debputy",
-        data=[propose_correct_text_quick_fix(n) for n in candidates],
+        data=DiagnosticData(
+            quickfixes=[propose_correct_text_quick_fix(n) for n in candidates]
+        ),
     )
     return diagnostic, corrected_key
 
@@ -922,7 +925,7 @@ def debputy_manifest_hover(
     )
     if km is None:
         _info("No keyword match")
-        return
+        return None
     parser, plugin_metadata, at_depth_idx = km
     _info(f"Match leaf parser {at_depth_idx}/{len(segments)} -- {parser.__class__}")
     hover_doc_text = resolve_hover_text(
@@ -1020,19 +1023,14 @@ def resolve_hover_text(
     return hover_doc_text
 
 
-def _hover_doc(ls: "LanguageServer", hover_doc_text: Optional[str]) -> Optional[Hover]:
+def _hover_doc(
+    ls: "DebputyLanguageServer", hover_doc_text: Optional[str]
+) -> Optional[Hover]:
     if hover_doc_text is None:
         return None
-    try:
-        supported_formats = ls.client_capabilities.text_document.hover.content_format
-    except AttributeError:
-        supported_formats = []
-    markup_kind = MarkupKind.Markdown
-    if markup_kind not in supported_formats:
-        markup_kind = MarkupKind.PlainText
     return Hover(
         contents=MarkupContent(
-            kind=markup_kind,
+            kind=ls.hover_markup_format(MarkupKind.Markdown, MarkupKind.PlainText),
             value=hover_doc_text,
         ),
     )
