@@ -70,8 +70,8 @@ _DOUBLE_ESCAPEES = re.compile(r'([\n`$"\\])')
 _REGULAR_ESCAPEES = re.compile(r'([\s!"$()*+#;<>?@\[\]\\`|~])')
 _PROFILE_GROUP_SPLIT = re.compile(r">\s+<")
 _DEFAULT_LOGGER: Optional[logging.Logger] = None
-_STDOUT_HANDLER: Optional[logging.StreamHandler] = None
-_STDERR_HANDLER: Optional[logging.StreamHandler] = None
+_STDOUT_HANDLER: Optional[logging.StreamHandler[Any]] = None
+_STDERR_HANDLER: Optional[logging.StreamHandler[Any]] = None
 
 
 def assume_not_none(x: Optional[T]) -> T:
@@ -763,7 +763,7 @@ def setup_logging(
             colorlog.ColoredFormatter(color_format, style="{", force_color=True)
         )
         logger = logging.getLogger()
-        if existing_stdout_handler is not None:
+        if existing_stderr_handler is not None:
             logger.removeHandler(existing_stderr_handler)
         _STDERR_HANDLER = stderr_handler
         logger.addHandler(stderr_handler)
@@ -771,7 +771,7 @@ def setup_logging(
         stderr_handler = logging.StreamHandler(sys.stderr)
         stderr_handler.setFormatter(logging.Formatter(colorless_format, style="{"))
         logger = logging.getLogger()
-        if existing_stdout_handler is not None:
+        if existing_stderr_handler is not None:
             logger.removeHandler(existing_stderr_handler)
         _STDERR_HANDLER = stderr_handler
         logger.addHandler(stderr_handler)
